#include <cstdio>

#include <iostream>
#include <set>
#include <vector>
#include <deque>
#include <map>

#include <petscsles.h>

#include <newmatio.h>

static char help[] = 
"Basic test for self scheduling algorithm FEM with PETSc";

//-------<*>-------<*>-------<*>-------<*>-------<*>------- 
#undef __FUNC__
#define __FUNC__ "main"
int main(int argc,char **args) {
  Vec b,u;
  Mat A;
  SLES sles;
  PC      pc;		/* preconditioner context */
  KSP     ksp;		/* Krylov subspace method context */

  // Number of nodes per element
  const int nel=3;
  // Space dimension
  const int ndim=2;
  
  PetscInitialize(&argc,&args,(char *)0,help);

  // Get MPI info
  int size,rank;
  MPI_Comm_size(PETSC_COMM_WORLD,&size);
  MPI_Comm_rank(PETSC_COMM_WORLD,&rank);

  // read nodes
  double x[ndim];
  int read;
  vector<double> xnod;
  vector<int> icone;
  FILE *fid = fopen("node.dat","r");
  assert(fid);
  while (true) {
    for (int k=0; k<ndim; k++) {
      read = fscanf(fid,"%lf",&x[k]);
      if (read == EOF) break;
    }
    if (read == EOF) break;
    xnod.push_back(x[0]);
    xnod.push_back(x[1]);
  }
  fclose(fid);
  int nnod=xnod.size()/ndim;
  PetscPrintf(PETSC_COMM_WORLD,"Read %d nodes.\n",nnod);

  // read elements
  int ix[nel];
  fid = fopen("icone.dat","r");
  assert(fid);
  while (true) {
    for (int k=0; k<nel; k++) {
      read = fscanf(fid,"%d",&ix[k]);
      if (read == EOF) break;
    }
    if (read == EOF) break;
    icone.push_back(ix[0]);
    icone.push_back(ix[1]);
    icone.push_back(ix[2]);
  }
  fclose(fid);
  
  int nelem=icone.size()/nel;
  PetscPrintf(PETSC_COMM_WORLD,"Read %d elements.\n",nelem);

  // read fixations stored as a map node -> value
  map<int,double> fixa;
  fid = fopen("fixa.dat","r");
  assert(fid);
  while (true) {
    int nod;
    double val;
    read = fscanf(fid,"%d %lf",&nod,&val);
    if (read == EOF) break;
    fixa[nod]=val;
  }
  fclose(fid);
  PetscPrintf(PETSC_COMM_WORLD,"Read %d fixations.\n",fixa.size());

  // Construct node to element pointer
  // n2e[j-1] is the set of elements connected to node 'j'
  vector< set<int> > n2e;
  n2e.resize(nnod);
  for (int j=0; j<nelem; j++) {
    for (int k=0; k<nel; k++) {
      int nodo = icone[j*nel+k];
      n2e[nodo-1].insert(j+1);
    }
  }

#if 0    // Output 'n2e' array if needed
  for (int j=0; j<nnod; j++) {
    set<int>::iterator k;
    PetscPrintf(PETSC_COMM_WORLD,"node %d, elements: ",j+1);
    for (k=n2e[j].begin(); k!=n2e[j].end(); k++) {
      PetscPrintf(PETSC_COMM_WORLD,"%d ",*k);
    }
    PetscPrintf(PETSC_COMM_WORLD,"\n");
  }
#endif

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
  // Simple partitioning algorithm 
  deque<int> q;
  vector<int> eord,id;
  eord.resize(nelem,0);
  id.resize(nnod,0);

  // Mark element 0 as belonging to processor 0
  q.push_back(1);
  eord[0]=-1;
  int order=0;

  
  while (q.size()>0) {
    // Pop an element from the queue
    int elem=q.front();
    q.pop_front();
    eord[elem-1]=++order;
    // Push all elements neighbor to this one in the queue
    for (int nod=0; nod<nel; nod++) {
      int node = icone[(elem-1)*nel+nod];
      set<int> &e = n2e[node-1];
      
      for (set<int>::iterator k=e.begin(); k!=e.end(); k++) {
        if (eord[*k-1]==0) {
          q.push_back(*k);
          eord[*k-1]=-1;
        }
      }
    }
  }
  q.clear();

  // Element partition. Put (approximmately)
  // nelem/size in each processor
  int *e_indx = new int[size+1];
  e_indx[0]=1;
  for (int p=0; p<size; p++) 
    e_indx[p+1] = e_indx[p]+(nelem/size) + (p<(nelem % size) ? 1 : 0);

  // Node partitioning. If a node is connected to an element 'j' then
  // put it in the processor where element 'j' belongs. 
  // In case the elements connected to the node belong to
  // different processors take any one of them. 
  int *id_indx = new int[size+2];
  for (int j=0; j<nnod; j++) {
    if (fixa.find(j+1) != fixa.end()) {
      id[j]=-(size+1);
    } else {
      set<int> &e = n2e[j];
      assert(e.size()>0);
      int order = eord[*e.begin()-1];
      for (int p=0; p<size; p++) {
        if (order >= e_indx[p] && order < e_indx[p+1]) {
          id[j] = -(p+1);
          break;
        }
      }
    }
  }

  // `id_indx[j-1]' is the dof associated to node 'j'
  int dof=0;
  id_indx[0]=0;
  if (size>1) 
    PetscPrintf(PETSC_COMM_WORLD,
		"dof distribution among processors\n");
  for (int p=0; p<=size; p++) {
    for (int j=0; j<nnod; j++) 
      if (id[j]==-(p+1)) id[j]=dof++;
    id_indx[p+1] = dof;
    if (p<size) {
      PetscPrintf(PETSC_COMM_WORLD,
                  "proc: %d, from %d to %d, total %d\n",
                  p,id_indx[p],id_indx[p+1],id_indx[p+1]-id_indx[p]);
    } else {
      PetscPrintf(PETSC_COMM_WORLD,
                  "fixed: from %d to %d, total %d\n",
                  id_indx[p],id_indx[p+1],id_indx[p+1]-id_indx[p]);
    }      
  }
  n2e.clear();
  
  int ierr;
  // Total number of unknowns (equations)
  int neq = id_indx[size];
  // Number of unknowns in this processor 
  int ndof_here = id_indx[rank+1]-id_indx[rank];
  // Creates a square MPI PETSc matrix
  ierr = MatCreateMPIAIJ(PETSC_COMM_WORLD,ndof_here,ndof_here,
			 neq,neq,0,
			 PETSC_NULL,0,PETSC_NULL,&A); CHKERRQ(ierr);
  // Creates PETSc vectors
  ierr = VecCreateMPI(PETSC_COMM_WORLD,
		      ndof_here,neq,&b); CHKERRQ(ierr);
  ierr = VecDuplicate(b,&u); CHKERRQ(ierr);
  double scal=0.;
  ierr =  VecSet(&scal,b);
  
  { // Compute element matrices and load them. 
    // Each processor computes the elements belonging to him. 
    
    // x12:= vector going from first to second node,
    // x13:= idem 1->3. 
    // x1:= coordinate of node 1
    // gN gradient of shape function
    ColumnVector x12(ndim),x13(ndim),x1(ndim),gN(ndim);
    // grad_N := matrix whose columns are gradients of interpolation
    // functions
    // ke:= element matrix for the Laplace operator
    Matrix grad_N(ndim,nel),ke(nel,nel);
    // area:= element area
    double area;
    for (int e=1; e<=nelem; e++) {
      int ord=eord[e-1];
      // skip if the element doesn't belong to this processor
      if (ord < e_indx[rank] || ord >= e_indx[rank+1]) continue;
      // indices for vertex nodes of this element
      int n1,n2,n3;
      n1 = icone[(e-1)*nel];
      n2 = icone[(e-1)*nel+1];
      n3 = icone[(e-1)*nel+2];
      x1 << &xnod[(n1-1)*ndim];
      x12 << &xnod[(n2-1)*ndim];
      x12 = x12 - x1;
      x13 << &xnod[(n3-1)*ndim];
      x13 = x13 - x1;
      // compute as vector product
      area = (x12(1)*x13(2)-x13(1)*x12(2))/2.;
      // gradients are proportional to the edge
      // vector rotated 90 degrees
      gN(1) = -x12(2); 
      gN(2) = +x12(1); 
      grad_N.Column(3) = gN/(2*area);
      gN(1) = +x13(2); 
      gN(2) = -x13(1); 
      grad_N.Column(2) = gN/(2*area);
      // last gradient can be computed from \sum_j grad_N_j = 0
      grad_N.Column(1) = -(grad_N.Column(2)+grad_N.Column(3));
      // Integrand is constant over element
      ke = grad_N.t() * grad_N * area;

      // Load matrix element on global matrix
      int nod1,nod2,eq1,eq2;
      for (int i1=1; i1<=nel; i1++) {
        nod1 = icone[(e-1)*nel+i1-1];
        if (fixa.find(nod1)!=fixa.end()) continue;
        eq1=id[nod1-1];
        for (int i2=1; i2<=nel; i2++) {
          nod2 = icone[(e-1)*nel+i2-1];
          eq2=id[nod2-1];
          if (fixa.find(nod2)!=fixa.end()) {
            // Fixed nodes contribute to the right hand side
            VecSetValue(b,eq1,-ke(i1,i2)*fixa[nod2],ADD_VALUES);
          } else {
            // Load on the global matrix
            MatSetValue(A,eq1,eq2,ke(i1,i2),ADD_VALUES);
          }
        }
      }
      
    }
    // Finish assembly. 
    ierr = VecAssemblyBegin(b); CHKERRQ(ierr);
    ierr = VecAssemblyEnd(b); CHKERRQ(ierr);
    
    ierr = MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY); 
    ierr = MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);       
  }

  // Create SLES and set options
  ierr = SLESCreate(PETSC_COMM_WORLD,&sles); CHKERRQ(ierr);
  ierr = SLESSetOperators(sles,A,A,
                          DIFFERENT_NONZERO_PATTERN); CHKERRQ(ierr);
  ierr = SLESGetKSP(sles,&ksp); CHKERRQ(ierr);
  ierr = SLESGetPC(sles,&pc); CHKERRQ(ierr);
  
  ierr = KSPSetType(ksp,KSPGMRES); CHKERRQ(ierr);
  // This only works for one processor only
  //  ierr = PCSetType(pc,PCLU); CHKERRQ(ierr);
  ierr = PCSetType(pc,PCJACOBI); CHKERRQ(ierr);
  ierr = KSPSetTolerances(ksp,1e-6,1e-6,1e3,100); CHKERRQ(ierr);
  int its;
  ierr = SLESSolve(sles,b,u,&its); 

  // Print vector on screen
  ierr =  VecView(u,PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr); 

  // Cleanup. Free memory
  delete[] e_indx;
  delete[] id_indx;
  PetscFinalize();
  
}
