//$Id: find-par.cpp,v 1.5 2006/05/20 15:05:03 mstorti Exp $
#include <cstdio>
#include <cmath>
#include <cassert>
#include <cstdlib>

#include <mpi.h>

// This is the `scalar' find. it is O(n*m).
// This could be improved by sorting `y' and
// using `binary-search'. If the element is not
// found use an invalid index (-1). 
void find(int *x,int n,int *y,int m,int *indx) {
  for (int j=0; j<n; j++) {
    indx[j] = -1;
    for (int k=0; k<m; k++) {
      if (x[j]==y[k]) {
	indx[j] = k;
	break;
      }
    }
  }
}

void find_par(int *x,int n,int *y,int m,int *indx,int root) {
  int  myrank, numprocs;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);

  // Bcast size of vectors and `y' vector.
  // `y' is allocated in the slaves and afterwards deleted. 
  MPI_Bcast(&n,1,MPI_INT,root,MPI_COMM_WORLD);
  MPI_Bcast(&m,1,MPI_INT,root,MPI_COMM_WORLD);
  if (myrank!=root) {
    y = new int[m];
  }

  // Bcast `y' vector
  MPI_Bcast(y,m,MPI_INT,root,MPI_COMM_WORLD);

  // Compute ranges, `cnts' and `displs'
  // They are allocated on `root' and
  // afterwards deleted. 
  int rest = n % numprocs;
  int *cnts, *displs;
  if (myrank==root) {
    cnts = new int[numprocs];
    displs = new int[numprocs+1];
    displs[0]=0;
    for (int j=0; j<numprocs; j++) {
      // cnts[j[ is either `(n/numprocs)' of `(n/numprocs)+1'
      // `displs' is computed as a `cum-sum' of `cnts'
      cnts[j] = (n/numprocs) + (j<rest? 1 : 0);
      displs[j+1] = displs[j] + cnts[j];
    }
  } 
  // My range to compute in `x' is `[x1,x2)'.
  // The range is computed bcasted even tot he root. 
  int x1 = myrank*(n/numprocs)+ (myrank<rest? myrank : rest);
  int x2 = x1 + (n/numprocs) + (myrank<rest? 1 : 0);
  int local_size = x2-x1;
  
  // These are allocated and deleted afterwards. 
  int *local_indx = new int[local_size];
  int *local_x = new int[local_size];

  // Scatter vector `x' to the slaves. 
  int ierr = MPI_Scatterv(x,cnts,displs,MPI_INT,
			  local_x,local_size,MPI_INT,root,MPI_COMM_WORLD);

  // Apply the scalar `find'
  find(local_x,local_size,y,m,local_indx);
#if 0
  for (int j=0; j<local_size; j++) 
    printf("[%d] j %d, x[j] %d, local_indx[j] %d, y[indx] %d\n",
	   myrank,j,local_x[j],local_indx[j],y[local_indx[j]]);
#endif

  // Gather all the indices
  ierr = MPI_Gatherv(local_indx,local_size,MPI_INT,
		     indx,cnts,displs,MPI_INT,root,MPI_COMM_WORLD);

  // Delete auxiliary vectors
  if (myrank==root) {
    delete[] cnts;
    delete[] displs;
  } else {
    delete[] y;
  }
  delete[] local_indx;
  delete[] local_x;
}

int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);

  int  myrank, numprocs;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);

  // Generate m random integers in the
  // range `10*m' in `y'. Pick out `n'
  // random from them. 
  int root=0, n=10000, m=100000;
  int *x, *y, *indx;
  if (myrank==root) {
    x = new int[n];
    indx = new int[n];
    y = new int[m];
    for (int j=0; j<m; j++)
      y[j] = rand() % (10*m);
    for (int j=0; j<n; j++) {
      int k = rand() % m;
      x[j] = y[k];
      // printf("j %d, x[j] %d\n",j,x[j]);
    }
  }

  // Call the parallel `find'
  find_par(x,n,y,m,indx,root);

  // Check that the `x' values are in the
  // computed indices. `bad' is the count
  // of erroneous computations (It should be null). 
  if (myrank==root) {
    int bad = 0;
    for (int j=0; j<n; j++) 
      bad += (x[j] != y[indx[j]]);
    printf("numprocs %d, checked %d of %d, bad %d\n",
	   numprocs,n,m,bad);

    delete[] x;
    delete[] y;
    delete[] indx;
  }
  MPI_Finalize();
}
