//  Copyright (C) 1999-2002, Mario Alberto Storti, Centro
//  Internacional de Metodos Numericos en Ingenieria
//  (CIMEC-Argentina), Universidad Nacional del Litoral
//  (UNL-Argentina), Consejo Nacional de Investigaciones Cientificas y
//  Tecnicas (UNL-Argentina).
//  
//  This program is  free software; you can redistribute  it and/or modify
//  it under the  terms of the GNU General Public  License as published by
//  the Free Software Foundation; either  version 2 of the License, or (at
//  your option) any later version.
//  
//  This program  is distributed in the  hope that it will  be useful, but
//  WITHOUT   ANY  WARRANTY;   without  even   the  implied   warranty  of
//  MERCHANTABILITY  or FITNESS  FOR A  PARTICULAR PURPOSE.   See  the GNU
//  General Public License for more details.
//  
//  You  should have received  a copy  of the  GNU General  Public License
//  along  with  this  program;  if   not,  write  to  the  Free  Software
//  Foundation, Inc.,  59 Temple Place, Suite 330,  Boston, MA 02111-1307,
//  USA.

// $Id: life.cpp,v 1.3 2004/09/13 11:37:08 mstorti Exp $
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <mpi.h>
#include <cstdio>
#include <cassert>
#include <error.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <cstring>
#include <vector>
#include <cmath>
#include <time.h>
#include <sys/time.h>

using namespace std;

double etime(timeval &x,timeval &y) {
  return double(y.tv_sec)-double(x.tv_sec)
    +(double(y.tv_usec)-double(x.tv_usec))/1e6;
}

int wait_from_console(char *s=NULL);

void go(int N, char *old0, char *old1, char *old2, char *new1);

// Simulates the `Game of Life' in parallel using MPI. 

/// This represents the board. 
class Board {
private:
  // N is the size of the board, so that it has N x N cells NN = N+2
  // is the size with two rows added at each end for simulating the
  // walls of the board.  We use a 0-based (C-style) indices i,j for
  // rows and columns.  Each processor owns a range of consecutive
  // rowns with indices i1<= i <i2. Currently the number of rows in
  // each processor is almost the same (no load balance). For
  // instance, if N=20 and size=2, then i1=0, i2=10 in processor 0 and
  // i1=10, i2=20 in processor 1.  So that each processor owns i2-i1
  // rows. We store a board that has one additional row at each end
  // (so called `ghost' rows), so that in fact at each processor we
  // store i2-i1+2 rows.  We define I1=i1-1 and I2=i2+1 so that the
  // rows stored at this processor are I1 <= i < I2. Ghost rows I1 and
  // I2-1 are computed at processors `myrank-1' and `myrank+1'. So
  // that they have to be transmitted at each time step with MPI.  The
  // flag `periodic' indicates whether we use periodic boundary
  // conditions or not (not used currently though!!).
  int N,NN,i1,i2,I1,I2,size,myrank,periodic;
  // Number of rows processed since last `reset_statistics()'
  int processed_rows;
  // Number of secods since last `reset_statistics()'
  double cum_time;
  // Number of secods idle (or communicating)
  double cum_idle_time;
  // Weights of processors
  vector<double> w;
  // Number of rows in each processor
  vector<int> rows_proc;
  // This arrays store the board. `board2' is a copy of the board for
  // when updating the board. 
  char **board,*row1,*row2;
  // returns the position in the board of cell j,k
  inline char *pos(int j, int k);
  void print_row(int k);
  // Frees resources
  void clear();
  // Read weights file from "weights.dat"
  void read_weights_file(const char *w);
  // Initializes the board. Generic
  void init(int N_,const char *wfile,vector<double> *w,int periodic);
public:
  // Initializes the board. Reads weights from file. 
  void init(int N_,const char *wfile=NULL,int periodic=0);
  // Initializes the board. A vector of weights is passed. 
  void init(int N_,const vector<double> weights,int periodic=0);
  // Sets the the state of a cell. 
  void setv(int j,int k,int val);
  // rescatter the board according to new weights
  void re_scatter(vector<double> &new_weight);
  // Prints the board
  void print();
  // Advances the board one generation.
  void advance();
  // Print statistics about estiamted processing rates
  void statistics(vector<double> &cum_time_v,
		  vector<double> &cum_idle_time_v,
		  vector<double> &processed_rows_v);
  // Reset statistics
  void reset_statistics();
  // Ctor.
  Board() { board = NULL; cum_time=0.; cum_idle_time = 0.; processed_rows=0; }
  // Dtor.
  ~Board();
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Board::setv(int j,int k,int val) {
  if (k>=i1 && k<i2) *pos(j,k) = val;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
inline char *Board::pos(int j, int k) { 
  assert(k<I2);
  return &board[(k-I1)][j+1];
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Board::read_weights_file(const char *opt_w) {
  // Read processors speed from file
  double r;
  char *line=NULL;
  size_t n=0;
  w.clear();
  w.resize(size);
  if (opt_w) {
    // if (myrank==0) printf("Reading weights file\n");
    FILE *fid = fopen("weights.dat","r");
    assert(fid);
    for (int p=0; p<size; p++) {
      int nread = getline(&line,&n,fid);
      assert(nread>0);
      nread = sscanf(line,"%lf",&r);
      assert(nread==1);
      w[p] = r;
    }
    fclose(fid);
    if (line) free(line);
  } else for (int p=0; p<size; p++) w[p] = 1.;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Board::print() {
  // This is done in parallel and may not work
  // correctly sometimes. 

  // Loop over processors. This is synchronized by the barrier below. 
#if 0
  for (int r=0; r<size; r++ ) {
    if (r==myrank) {
      for (int k=i1; k<i2; k++) {
	char *p = pos(0,k), *pe=p+N;
	while (p<pe) {
	  putc((*p++ ? '*' : '.'),stdout);
	  putc(' ',stdout);
	}
	putc('\n',stdout);
      }
      fflush(stdout);
    }
    MPI_Barrier(MPI_COMM_WORLD);
    if (myrank==0) fflush(stdout);
  }
  if (myrank==0) printf("\n");
#endif
  char *buffer = new char[N*N];
  char *buffer2 = new char[N*N];
  // Initialize to 0
  for (int j=0; j<N; j++)
    for (int k=0; k<N; k++) buffer[j*N+k] = 0;
  // Each processor sets their corresponding cells
  for (int j=i1; j<i2; j++) 
    for (int k=0; k<N; k++)
      buffer[k*N+j] = *pos(k,j);
  MPI_Reduce(buffer,buffer2,N*N,MPI_CHAR,MPI_LOR,0,MPI_COMM_WORLD);
  if (!myrank) {
    for (int j=0; j<N; j++) {
      for (int k=0; k<N; k++) {
	putchar((buffer2[j*N+k] ? '*' : '.'));
	putchar(' ');
      }
      putchar('\n');
    }
  }
  delete[] buffer;
  delete[] buffer2;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Board::print_row(int k) {
  char *p = pos(0,k), *pe = p+N;
  printf("row %3d -> ",k);
  while (p < pe) {
    putc((*p++ ? '*' : '.'),stdout);
    putc(' ',stdout);
  }
  putc('\n',stdout);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
// Computes next generation 
void Board::advance() {
  
  double t1,t2;
  t1 = MPI_Wtime();
  
#define SEND(row,proc)    \
      MPI_Send(pos(0,(row)),N,MPI_CHAR,(proc),0,MPI_COMM_WORLD)
#define RECEIVE(row,proc) \
    MPI_Recv(pos(0,(row)),N,MPI_CHAR,(proc),MPI_ANY_TAG,MPI_COMM_WORLD,&stat)

  struct timeval start, end;
  // First stage 2j -> 2j+1
  MPI_Status stat;
  if (myrank % 2 == 0 ) {
    if (myrank<size-1) {
      SEND(i2-1,myrank+1);
      RECEIVE(I2-1,myrank+1);
    }
    if (myrank>0) {
      SEND(i1,myrank-1);
      RECEIVE(I1,myrank-1);
    }
  } else {
    RECEIVE(I1,myrank-1);
    SEND(i1,myrank-1);
    if (myrank<size-1) {
      RECEIVE(I2-1,myrank+1);
      SEND(i2-1,myrank+1);
    }
  }

  t2 = MPI_Wtime();
  cum_idle_time += t2-t1;
  t1 = t2;

  // Initialize row2
  { char *q = row2, *qe = q+NN;
  while (q<qe) *q++ = 0;
  }

  // Loop over rows in this processor (not included the `ghost' ones
  // (I1 and I2-1)
  int k; char *tmp;
  for (k=i1; k<i2; k++) {
    go(N, pos(-1,k-1), pos(-1,k), pos(-1,k+1), row1);
    tmp = board[k-1-I1];
    board[k-1-I1] = row2;
    row2 = row1;
    row1 = tmp;
  }
  tmp = board[k-1-I1];
  board[k-1-I1] = row2;
  row2 = row1;
  row1 = tmp;

  cum_time += MPI_Wtime()-t1;
  processed_rows += i2-i1;

}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Board::statistics(vector<double> &cum_time_v,
		       vector<double> &cum_idle_time_v,
		       vector<double> &processed_rows_v) {

  cum_time_v.resize(size);
  cum_idle_time_v.resize(size);
  processed_rows_v.resize(size);

  MPI_Allgather(&cum_time,1,MPI_DOUBLE,&*cum_time_v.begin(),
		1,MPI_DOUBLE,MPI_COMM_WORLD);
  MPI_Allgather(&cum_idle_time,1,MPI_DOUBLE,&*cum_idle_time_v.begin(),
		1,MPI_DOUBLE,MPI_COMM_WORLD);
  double pr = double(processed_rows);
  MPI_Allgather(&pr,1,MPI_DOUBLE,&*processed_rows_v.begin(),
		1,MPI_DOUBLE,MPI_COMM_WORLD);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Board::reset_statistics() {
  cum_time = 0.;
  cum_idle_time = 0.;
  processed_rows = 0;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void pv(int myrank,int size,vector<int> &v,char *s=NULL) {
  int n = v.size();
  vector<int> buff(size*n);
  MPI_Gather(&*v.begin(),n,MPI_INT,&*buff.begin(),n,MPI_INT,0,MPI_COMM_WORLD);
  if (!myrank) {
    if (s) printf("%s",s);
    for (int p=0; p<size; p++) {
      printf("[%d] ",p);
      for (int k=0; k<v.size(); k++) printf("%d ",buff[p*n+k]);
      printf("\n");
    }
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Board::init(int N,vector<double> ww,int per) {
  init(N,NULL,&ww,per);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Board::init(int N,const char *wfile,int per) {
  init(N,wfile,NULL,per);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
// Initialize board
void Board::init(int N_,const char *wfile,vector<double> *ww,int per) {

  // size = number of processors
  MPI_Comm_size(MPI_COMM_WORLD,&size);
  // my rank in the row of processors
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);

  // flags whther periodic b.c.'s are used or not
  periodic = per;
  // size of board
  N = N_;

  if (ww) w = *ww;
  else read_weights_file(wfile);

  if (w.size()>0 && w.size()<size ) {
    if (myrank==0) printf("number of weights in \"weights\" "
			  "file doesn't match number of procs.\n");
    MPI_Finalize();
    exit(1);
  } 
  if (w.size()==0) w.resize(size,1.);
  rows_proc.resize(size);    //L-part-b
  double sum_w = 0., carry=0., a;
  for (int p=0; p<size; p++) sum_w += w[p];
  int j = 0;
  double tol = 1e-8;
  for (int p=0; p<size; p++) {
    w[p] /= sum_w;
    a = w[p]*N + carry + tol;
    rows_proc[p] = int(floor(a));
    carry = a - rows_proc[p] - tol;
    if (p==myrank) {
      i1 = j;
      i2 = i1 + rows_proc[p];
    }
    j += rows_proc[p];
  }
  assert(j==N);
  assert(fabs(carry) < tol);
  
  I1=i1-1;
  I2=i2+1;
                           //L-part-e
  // Alloc memory
  NN = N+2;			// number of stored rows (include ghosts)
  board  = new (char *)[I2-I1]; // allocate memory for boards
  for (int k=0; k<I2-I1; k++) board[k] = new char[NN];
  row1 = new char[NN];
  row2 = new char[NN];
  
  // Initialize board
  for (int k=I1; k<I2; k++) {
    char *p = pos(-1,k), *pe = p+N+2;
    while (p < pe) *p++ = 0;
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void Board::re_scatter(vector<double> &new_weight) {

  // Rows in processor `p', old/new partitioning are
  // send_ptr[p] <= k < send_ptr[p+1]
  // recv_ptr[p] <= k < recv_ptr[p+1]
  vector<int> send_ptr(size+1), recv_ptr(size+1);

  // Define old partitioning
  send_ptr[0]=0;
  for (int p=0; p<size; p++) send_ptr[p+1] = send_ptr[p] + rows_proc[p];

  // Buffer for storing old rows
  char *send_buff = new char[rows_proc[myrank]*N];

  // Store old row values
  for (int k=i1; k<i2; k++) {
    char *p,*pe,*q;
    p = pos(0,k);
    pe = p+N;
    q = send_buff+(k-i1)*N;
    while (p<pe) { *q++ = *p++; }
  }

  // Reinitialize the board
  clear();
  assert(w.size() == new_weight.size());
  init(N,new_weight,periodic);

  // Buffer for storing new rows
  char *recv_buff = new char[rows_proc[myrank]*N];

  // Define new partitioning
  recv_ptr[0]=0;
  for (int p=0; p<size; p++) recv_ptr[p+1] = recv_ptr[p] + rows_proc[p];

  // Use MPI_Alltoallv

  vector<int> send_counts(size), sdispls(size+1);
  vector<int> recv_counts(size), rdispls(size+1);
  
  // Compute send_counts and sdispls
  // [k1,k2) was the range previously here
  int k1 = send_ptr[myrank];
  int k2 = send_ptr[myrank+1];
  for (int p=0; p<size+1; p++) {
    int q = recv_ptr[p];
    if (q<k1) q = k1;
    if (q>k2) q = k2;
    q -= k1;
    sdispls[p] = q*N;
    if (p>0) send_counts[p-1] = sdispls[p]-sdispls[p-1];
  }

  // Compute recv_counts and rdispls
  // [i1,i2) is the range currently here
  for (int p=0; p<size+1; p++) {
    int q = send_ptr[p];
    if (q<i1) q = i1;
    if (q>i2) q = i2;
    q -= i1;
    rdispls[p] = q*N;
    if (p>0) recv_counts[p-1] = rdispls[p]-rdispls[p-1];
  }

  // wait_from_console("Before MPI_Alltoallv\n");
  int ierr = MPI_Alltoallv(send_buff,&*send_counts.begin(),
			   &*sdispls.begin(),MPI_CHAR,
			   recv_buff,&*recv_counts.begin(),
			   &*rdispls.begin(),MPI_CHAR,
			   MPI_COMM_WORLD);
  // wait_from_console("After MPI_Alltoallv\n");
  assert(!ierr);

  // Copy new rows to board
  for (int k=i1; k<i2; k++) {
    char *p,*pe,*q;
    p = pos(0,k);
    pe = p+N;
    q = recv_buff+(k-i1)*N;
    while (p<pe) { *p++ = *q++; }
  }

  delete[] send_buff;
  delete[] recv_buff;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
// Frees resources
void Board::clear() {
  for (int k=0; k<I2-I1; k++) 
    delete[] board[k];
  delete[] board;
  board=NULL;
  delete[] row1;
  row1 = NULL;
  delete[] row2;
  row2 = NULL;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
// Destructor
Board::~Board() { clear(); }

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
// Main program
int main(int argc,char **args) {
  int size,myrank;
  char *pattern = NULL;
  char *file = NULL;

  // Initializes MPI
  MPI_Init(&argc,&args);
  // size = number of processors
  MPI_Comm_size(MPI_COMM_WORLD,&size);
  // my rank in the row of processors
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);

  int N=20,nstep=30,opt_q=0,opt_v=0,nstep_rs;
  double opt_r=0.;
  char *opt_w = NULL;
  char c;

  while ((c = getopt(argc, args, "N:s:hqp:vw:S:r:f:")) != -1) {
    switch (c) {
    case 'h':
      if (myrank==0) {
	printf(" usage: $ mpirun [OPTIONS] life [life-options]\n\n"
	       "MPI options: -np <number-of-processors>\n"
	       "             -machinefile <machine-file>\n\n"
	       "Life options:  -N <number of columns/rows>\n"
	       "               -s <numer of time steps>\n"
	       "               -S <numer of time steps between each re_scatter>\n"
	       "               -r <rand-frac> # 0<rf< 1, rescatter with this degree of random\n"
	       "               -q             # quiet (no output)\n"
	       "               -v             # verbose (print board at each step)\n"
	       "               -p <pattern>   # initial pattern, may be ('cross',line10)\n"
	       "               -f <file-pattern> # read initial pattern from file\n"
	       "               -w <w-file>    # read weights from <w-file>\n"
	       "               -h             # print this help\n"
	       );
      }
      MPI_Finalize();
      exit(0);
    case 'N':
      sscanf(optarg,"%d",&N);
      break;
    case 'w':
      opt_w = new char[strlen(optarg)+1];
      strcpy(opt_w,optarg);
      break;
    case 's':
      sscanf(optarg,"%d",&nstep);
      break;
    case 'q':
      opt_q = 1;
      break;
    case 'v':
      opt_v = 1;
      break;
    case 'p':
      pattern = new char[strlen(optarg)+1];
      strcpy(pattern,optarg);
      break;
    case 'f':
      file = new char[strlen(optarg)+1];
      strcpy(file,optarg);
      break;
    case 'S':
      sscanf(optarg,"%d",&nstep_rs);
      break;
    case 'r':
      opt_r = 1.;
      int nread;
      if (optarg) nread = sscanf(optarg,"%lf",&opt_r);
      if (nread!=1) opt_r = 1.;
      break;
    case '?':
      if (isprint (optopt))
	fprintf (stderr, "Unknown option `-%c'.\n", optopt);
      else
	fprintf (stderr,
		 "Unknown option character `\\x%x'.\n",
		 optopt);
      return 1;
    default:
      if (isprint (optopt))
	fprintf (stderr, "Unknown option `-%c'.\n", optopt);
      else
	fprintf (stderr,
		 "Unknown option character `\\x%x'.\n",
		 optopt);
      abort ();
    }
  }

  if (nstep_rs<0) nstep_rs = INT_MAX;
  /// time related quantities
  double start, end;

  // Defines and initializes the board
  Board board;
  board.init(N,opt_w);

  if (!file && !pattern) pattern = "cross";

  if (pattern && file) {
    error(1,0,"Can't use -p and -f options at the same time\n");
  }

  if (file) {
    FILE *fid = fopen(file,"r");
    assert(fid);
    for (int j=0; j<N; j++) {
      char *line = NULL;
      size_t len = 0;
      int nread = getline(&line,&len,fid);
      assert(nread==N+1);
      for (int k=0; k<N; k++) {
	switch (*(line+k)) {
	case '*':
	  board.setv(j,k,1);
	  break;
	case '-':
	  board.setv(j,k,0);
	  break;
	default:
	  error(1,0,"Invalid char \"%c\"",*(line+j));
	  break;
	}
      }
    }
    fclose(fid);
  } else {
    if (!strcmp(pattern,"cross")) {
      for (int k=0; k<N; k++) {
	board.setv(N/2,k,1);
	board.setv(k,N/2,1);
      }
    } else if (!strcmp(pattern,"line10")) {
      assert(N>20);
      for (int k=-5; k<5; k++) board.setv(N/2,N/2+k,1);
    } else if (!strcmp(pattern,"block")) {
      assert(N>3);
      board.setv(N/2  ,N/2  ,1);
      board.setv(N/2  ,N/2-1,1);
      board.setv(N/2-1,N/2  ,1);
      board.setv(N/2-1,N/2-1,1);
    } else {
      if (myrank==0) 
	printf("Not known pattern \"%s\"\n",pattern);
      MPI_Finalize();
      exit(0);
    }
  }

  if (!opt_q) board.print();
  start = MPI_Wtime();

  vector<double> cum_time_v, cum_idle_time_v, processed_rows_v;
  for (int k=0; k<nstep; k++) {
    board.advance();
    if (opt_v) {
      if (myrank==0) printf("-- Generation %d:\n",k+1); 
      board.print();
    }

    int ll = 0;
    if (!((k+1) % nstep_rs)) {
      double tstart=MPI_Wtime();
      // board.print();
      vector<double> neww(size);
      board.statistics(cum_time_v,cum_idle_time_v,processed_rows_v);

      if (myrank==0) printf("re_scatter %d, weights: ",ll);
      double sumd=0;
      for (int p=0; p<size; p++) {
	double d = double(processed_rows_v[p])/cum_time_v[p];
	d *= (1.-opt_r) + opt_r*double(rand())/double(RAND_MAX);
	neww[p] = d;
	sumd += d;
      }
      sumd /= size;
      for (int p=0; p<size; p++) {
	neww[p] /= sumd;
	if (myrank==0) printf("%f ",neww[p]);
      }
      if (myrank==0) printf("\n");

      board.re_scatter(neww);
 
      if (!myrank) {
	double prow=0, prowi=0,mint,maxt;
	for (int j=0; j<size; j++) {
	  double ct = cum_time_v[j],
	    cit = cum_idle_time_v[j],
	    idle = cit/(cit+ct)*100.;
	  prow += double(processed_rows_v[j]);
	  prowi += double(processed_rows_v[j])/cum_time_v[j]*cum_idle_time_v[j];
	  double t = cum_idle_time_v[j]+cum_time_v[j];
	  if (j==0) {
	    maxt = mint = t;
	  } else {
	    if (t>maxt) maxt = t;
	    if (t<mint) mint = t;
	  }
	  printf("[%d] proc %d rows, t %fs, t_comm %fs, rate %f Mcells/s, idle %5.2f%%\n",
		 j,int(processed_rows_v[j]),cum_time_v[j],cum_idle_time_v[j],
		 processed_rows_v[j]*N/cum_time_v[j]/1e6,idle);
	}
	printf("min t %fs, max t %fs\n",mint,maxt);
	printf("rate %f, max rate %f [Mcells/s]\n",
	       prow/maxt/1e6,(prow+prowi)/maxt/1e6);
      }
      board.reset_statistics();
      if (myrank==0) printf("re_scatter: %fsecs\n",MPI_Wtime()-tstart);
    }
  }
    
  if (myrank==0) {
    end = MPI_Wtime();
    double elapsed = end-start;
    double rate = double(N)*double(N)*double(nstep)/elapsed/1e6;
    printf("N %d, steps %d, elapsed %f, rate: %f Mcell/sec\n",
	   N,nstep,elapsed,rate);
  }

  if (!opt_q && !opt_v) {
    if (myrank==0) printf("-- Final board:\n"); 
    board.print();
  }
  MPI_Finalize();
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void go(int N, char *old0, char *old1, char *old2, char *new1) {
  register char 
    *p = old1+1,
    *pe = p+N, 
    *pNE = old2+2, 
    *pSE = old0+2,
    *pnew = new1+1;
  char SW, S, SE, W, O, E, NW, NN, NE;
  SW = *(pSE-2);
  S  = *(pSE-1);
  SE = *(pSE  );
    
  W = *(p-1);
  O = *(p  );
  E = *(p+1);

  NW = *(pNE-2);
  NN  = *(pNE-1);
  NE = *(pNE  );
    
  // Cells computed in this processor in this row have 
  // p <= address < pe
  register int alive;
  register int ali_W = SW + W + NW;
  register int ali_C = NN + O + S;
  register int ali_E;
  register char *pE = p+1, *pEe = pe+1;

  while (p < pe) {
    ali_E = SE + E + NE;
    // Neighbors at positions one row below
    alive = ali_W + ali_C + ali_E - O;
      
    if (!alive) {		// We treat this case before, perhaps
      *pnew = 0;			// we gain some acceleration. Cell
				// becomes dead
    } else if (alive==3) {
      *pnew = 1;			// Cell becomes alive
    } else if (alive!=2) {
      *pnew = 0;			// Cell becomes dead
    } else *pnew = *p;		// cell remains in its state
    O = E;
    ali_W = ali_C;
    ali_C = ali_E;
    pSE++; pNE++; pnew++; p++;	// update positions in board to East 
    SE = *pSE;
    NE = *pNE;
    E = *(p+1);
  }
}
