// $Id: rescatter.cpp,v 1.9 2005/04/29 15:43:11 mstorti Exp $
#include <vector>
#include <cstdio>
#include <cassert>
#include <cmath>
#include <mpi.h>

using namespace std;

void print_par(vector<int> &buff,const char *s=NULL) { //L-print-par-b
  int  myrank, size;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  // reception buffer in master, receive counts
  // and displacements
  vector<int> rbuff, recvcnts(size,0), displs(size,0);

  // Each procesor send its size
  int sendcnt = buff.size();
  MPI_Gather(&sendcnt,1,MPI_INT,
             &recvcnts[0],1,MPI_INT,0,MPI_COMM_WORLD);
  if (!myrank) {
    // Resize reception buffer and
    // compute displs[] in master
    int rsize = 0;
    for (int j=0; j<size; j++) rsize += recvcnts[j];
    rbuff.resize(rsize);
    displs[0] = 0;
    for (int j=1; j<size; j++) 
      displs[j] = displs[j-1] + recvcnts[j-1];
    
  }
  // Do the gather
  MPI_Gatherv(&buff[0],sendcnt,MPI_INT,
	      &rbuff[0],&recvcnts[0],&displs[0],MPI_INT,
	      0,MPI_COMM_WORLD);
  if (!myrank) {
    // Print all buffers in master
    if (s) printf("%s",s);
    for (int j=0; j<size; j++) {
      printf("in proc [%d]: ",j);
      for (int k=0; k<recvcnts[j]; k++) {
	int ptr = displs[j]+k;
	printf("%d ",rbuff[ptr]);
      }
      printf("\n");
    }
  }
}
                    //L-print-par-e
int switch_next(int elem,int size,void *data) {
  int key =*(int *)data;
  return (elem - key % 2) % size;
}

int rotate_elems(int elem,int size, void *data) {
  int key =*(int *)data;
  return (elem - key) % size;
}

int scramble(int elem,int size,void *data) {
  int key =*(int *)data;
  return (elem*key) % size;
}

int allproc(int elem,int size,void *data) {
  int key =*(int *)data;
  return key % size;
}

int random_scatter(int elem,int size,void *data) {
  vector<int> &random_key = *(vector<int> *)data;
  int proc = elem % size;
  proc = random_key[proc];
  return  proc;
}

typedef int (*proc_fun)(int x,int size,void *data);

void re_scatter(vector<int> &buff,int size, //L-code-b
	    int myrank,proc_fun proc,void *data) {
  vector<int> sendcnts(size,0);
  int N = buff.size();
  // Check how many elements should go to
  // each processor
  for (int j=0; j<N; j++) {
    int p = proc(buff[j],size,data);
    sendcnts[p]++;
  }

  // Dimension buffers and ptr vectors
  vector<int> sbuff(N);
  vector<int> sdispls(size);
  vector<int> recvcnts(size);
  vector<int> rdispls(size);
  sdispls[0] = 0;
  for (int j=1; j<size; j++) 
    sdispls[j] = sdispls[j-1] + sendcnts[j-1];

  // Reorder by processor from buff to sbuff
  for (int j=0; j<N; j++) {
    int p = proc(buff[j],size,data);
    int pos = sdispls[p];
    sbuff[pos] = buff[j];
    sdispls[p]++;
  }
  // Use `Alltoall' for scattering the dimensions
  // of the buffers to be received
  MPI_Alltoall(&sendcnts[0],1,MPI_INT,
	       &recvcnts[0],1,MPI_INT,MPI_COMM_WORLD);

  // Compute the `send' and `recv' displacements. 
  rdispls[0] = 0;
  sdispls[0] = 0;
  for (int j=1; j<size; j++) {
    rdispls[j] = rdispls[j-1] + recvcnts[j-1];
    sdispls[j] = sdispls[j-1] + sendcnts[j-1];
  }

  // Dimension the receive size
  int rsize = 0;
  for (int j=0; j<size; j++) rsize += recvcnts[j];
  buff.resize(rsize);

  // Do the scatter. 
  MPI_Alltoallv(&sbuff[0],&sendcnts[0],&sdispls[0],MPI_INT,
		&buff[0],&recvcnts[0],&rdispls[0],MPI_INT,
		MPI_COMM_WORLD);
}
                                     //L-code-e
int main(int argc, char **argv) {
  
  MPI_Init(&argc,&argv);

  int  myrank, size, ierr;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  // srand(time(NULL));

  int N = 10, M=1000;

#if 0
  int rest = N % size;
  int Nhere = N/size + (myrank<rest ? 1 : 0);
  vector<int> buff(Nhere);
  for (int j=0; j<Nhere; j++) 
    buff[j] = rand() % M;
#else
  vector<int> buff;
  if (!myrank) {
    buff.resize(N);
    for (int j=0; j<N; j++) 
      buff[j] = rand() % M;
  }
#endif

  int s;
  print_par(buff,"initial:\n");

  for (int k=0; k<2*size; k++) {
    void *data = &k;
    proc_fun f;
#if 0
    f = key_rotate;
#elif 0
    f = scramble;
#elif 0
    f = allproc;
#elif 1
    f = rotate_elems;
#elif 0
    f = random_scatter;
    vector<int> random_key(size);
    if (!myrank) {
      for (int j=0; j<size; j++) {
	int proc = rand() % size;
	random_key[j] = proc;
	if (!proc) printf("proc %d scattered to 0\n",j);
      }
    }
    MPI_Bcast(&random_key[0],size,MPI_INT,0,MPI_COMM_WORLD);
    data = &random_key;
#else
#error Not defined scatter function
#endif

    re_scatter(buff,size,myrank,f,data);
    print_par(buff,"after rescatter: -------\n");
  }
  
  MPI_Finalize();  
}
