// $Id: rich.cpp,v 1.2 2004/10/24 21:19:15 mstorti Exp $
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
//  Solves Ax=b using Richardson
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
#include <cstdio>
#include <cassert>
#include <cmath>
#include <cfloat>
#include <cstring>
#include <vector>
#include <algorithm>
#include <omp.h>

using namespace std;

int modulo(int k, int n, int *div=NULL) {
  int m = k % n;
  int d = k / n;
  if (m<0) {
    m += n;
    d -= 1;
  }
  if (div) *div = d;
  return m;
}

void mat_vec(vector<int> &I,vector<int > &J,
             vector<double> &AIJ, vector<double> &X, 
             vector<double> &Y) {
  int nc = AIJ.size(), N=X.size();
  
  for (int i=0; i<N; i++) 
    Y[i] = 0.0;

  for (int k=0; k<nc; k++) {
    int i = I[k];
    int j = J[k];
    double a = AIJ[k];
    Y[i] += a*X[j];
  }
}

int layers=2;

void gen_matrix(int N,vector<int> &I,vector<int > &J,
                vector<double> &AIJ, vector<double> &X, 
                vector<double> &B) {
  double sigma = 0.5;
  int nngbr = 2*layers+1, nc=nngbr*N;
  I.resize(nc);
  J.resize(nc);
  AIJ.resize(nc);

  X.resize(N);
  B.resize(N);

  int indx=0;
  for (int i=0; i<N; i++) {
    X[i] = i<N/2;
    for (int j=-layers; j<=layers; j++) {
      assert(indx<nc);
      I[indx] = i;
      J[indx] = modulo(i-j,N);
      AIJ[indx] = (j? -1.0/layers : 2.0+sigma);
      indx++;
    }
  }
  assert(indx==nc);

#if 1
  // Random-shuffle the entries of the IJ-A matrix
  vector<int> perm(nc), Ic(nc),Jc(nc);
  vector<double> AIJc(nc);
  for (int j=0; j<nc; j++)
    perm[j] = j;
  random_shuffle(perm.begin(),perm.end());
  for (int j=0; j<nc; j++) {
    int jj = perm[j];
    Ic[j] = I[jj];
    Jc[j] = J[jj];
    AIJc[j] = AIJ[jj];
  }
  I = Ic;
  J = Jc;
  AIJ = AIJc;
#endif

  mat_vec(I,J,AIJ,X,B);
}

int main(int argc, char **argv) {
  
  vector<int> I,J;
  vector<double> AIJ, Xex, X, B, Y;
  int N=10, itmax = 100;

  gen_matrix(N,I,J,AIJ,Xex,B);
  int nc = AIJ.size();

#if 1
  for (int i=0; i<N; i++) 
    printf("i %d, B %f, X %f\n",i,B[i],Xex[i]);
  
  for (int k=0; k<nc; k++) 
    printf(" A(%d,%d) = %f\n",I[k],J[k],AIJ[k]);
#endif

#if 1
  double omega = 0.4, tol=1e-2, norm, norm0;
  int iter;
  Y.resize(N);
  X.resize(N,0.0);
  double start = omp_get_wtime();
  double rnorm;
  for (iter=0; iter<itmax; iter++) {
    rnorm=0.0;
    mat_vec(I,J,AIJ,X,Y);
    for (int i=0; i<N; i++) {
      double Ri = B[i] - Y[i];
      rnorm += Ri*Ri;
      X[i] += omega*Ri;
    }
    rnorm = sqrt(rnorm);
    printf("iters %d, norm %g\n",iter-1,rnorm);
  }
  double erro=0.0, ee;
  for (int j=0; j<N; j++) {
    ee = X[j]-Xex[j];
    erro += ee*ee;
  }
  erro = sqrt(erro);
  printf("iter %d, norm %g, erro %g\n",iter,rnorm,erro);
  double elapsed = omp_get_wtime()-start;
  printf("elapsed %f.2secs, rate %g Gflop/sec\n",elapsed,
         2.0*(2.0*layers+1)*double(nc)*double(iter)/elapsed*1e-9);
#endif
  return 0;
}
