/*$Id curspar-1.0.0-15-gabee420 Thu Jun 14 00:46:44 2007 -0300$*/

static char help[] = 
  "Newton's method to solve a combustion-like 1D problem.\n";

#include "petscsnes.h"

/* 
   User-defined routines
*/
extern int resfun(SNES,Vec,Vec,void*);
extern int jacfun(SNES,Vec,Mat*,Mat*,MatStructure*,void*);

struct SnesCtx {
  int N;
  double k,c,h;
};

#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc,char **argv)
{
  SNES         snes;         /* nonlinear solver context */
  SLES         sles;         /* linear solver context */
  PC           pc;           /* preconditioner context */
  KSP          ksp;          /* Krylov subspace method context */
  Vec          x,r;         /* solution, residual vectors */
  Mat          J;            /* Jacobian matrix */
  int          ierr,its,size;
  PetscScalar  pfive = .5,*xx;
  PetscTruth   flg;
  SnesCtx ctx;
  int N = 10;
  ctx.N = N;
  ctx.k = 0.1;
  ctx.c = 1;
  ctx.h = 1.0/N;

  PetscInitialize(&argc,&argv,(char *)0,help);
  ierr = MPI_Comm_size(PETSC_COMM_WORLD,&size);CHKERRQ(ierr);
  if (size != 1) SETERRQ(1,"This is a uniprocessor example only!");

  // Create nonlinear solver context
  ierr = SNESCreate(PETSC_COMM_WORLD,&snes); CHKERRQ(ierr);
  ierr = SNESSetType(snes,SNESLS); CHKERRQ(ierr);

  // Create matrix and vector data structures;
  // set corresponding routines

  // Create vectors for solution and nonlinear function
  ierr = VecCreateSeq(PETSC_COMM_SELF,N+1,&x);CHKERRQ(ierr);
  ierr = VecDuplicate(x,&r); CHKERRQ(ierr);
  double scal = 0.9;
  ierr = VecSet(&scal,x); CHKERRQ(ierr);
  
  ierr = MatCreateMPIAIJ(PETSC_COMM_SELF,PETSC_DECIDE,
                         PETSC_DECIDE,N+1,N+1,
                         1,NULL,0,NULL,&J);CHKERRQ(ierr);

  ierr = SNESSetFunction(snes,r,resfun,&ctx); CHKERRQ(ierr);
  ierr = SNESSetJacobian(snes,J,J,jacfun,&ctx); CHKERRQ(ierr);

  ierr = SNESSolve(snes,x,&its);CHKERRQ(ierr);
  Vec f;
  ierr = VecView(x,PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);
  ierr = SNESGetFunction(snes,&f,0,0);CHKERRQ(ierr);
  double rnorm;
  ierr = VecNorm(r,NORM_2,&rnorm);

  ierr = PetscPrintf(PETSC_COMM_SELF,
                     "number of Newton iterations = "
                     "%d, norm res %g\n",
                     its,rnorm);CHKERRQ(ierr);

  ierr = VecDestroy(x);CHKERRQ(ierr);
  ierr = VecDestroy(r);CHKERRQ(ierr);
  ierr = SNESDestroy(snes);CHKERRQ(ierr);

  ierr = PetscFinalize();CHKERRQ(ierr);
  return 0;
}

/* ------------------------------------------------------------------- */
#undef __FUNCT__
#define __FUNCT__ "resfun"
int resfun(SNES snes,Vec x,Vec f,void *data)
{
  double *xx,*ff;
  SnesCtx &ctx = *(SnesCtx *)data;
  int ierr;
  double h = ctx.h;

  ierr = VecGetArray(x,&xx);CHKERRQ(ierr);
  ierr = VecGetArray(f,&ff);CHKERRQ(ierr);

  ff[0] = xx[0];
  ff[ctx.N] = xx[ctx.N];

  for (int j=1; j<ctx.N; j++) {
    double xxx = xx[j];
    ff[j] = xxx*(0.5-xxx)*(1.0-xxx);
    ff[j] += ctx.k*(-xx[j+1]+2.0*xx[j+1]-xx[j-1])/(h*h);
  }

  ierr = VecRestoreArray(x,&xx);CHKERRQ(ierr);
  ierr = VecRestoreArray(f,&ff);CHKERRQ(ierr);
  return 0;
}

/* ------------------------------------------------------------------- */
#undef __FUNCT__
#define __FUNCT__ "resfun"
int jacfun(SNES snes,Vec x,Mat* jac,Mat* jac1,
           MatStructure *flag,void *data) {
  double *xx, A;
  SnesCtx &ctx = *(SnesCtx *)data;
  int ierr, j;

  ierr = VecGetArray(x,&xx);CHKERRQ(ierr);
  ierr = MatZeroEntries(*jac);

  j=0; A = 1;
  ierr = MatSetValues(*jac,1,&j,1,&j,&A,
                      INSERT_VALUES); CHKERRQ(ierr);

  j=ctx.N; A = 1;
  ierr = MatSetValues(*jac,1,&j,1,&j,&A,
                      INSERT_VALUES); CHKERRQ(ierr);

  for (j=1; j<ctx.N; j++) {
    double xxx = xx[j];
    A = (0.5-xxx)*(1.0-xxx) - xxx*(1.0-xxx) - xxx*(0.5-xxx);
    ierr = MatSetValues(*jac,1,&j,1,&j,&A,INSERT_VALUES); 
    CHKERRQ(ierr);
  }
  ierr = MatAssemblyBegin(*jac,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(*jac,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = VecRestoreArray(x,&xx);CHKERRQ(ierr);
  return 0;
}
